<?php

namespace Drupal\auctioneer;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\Display\EntityDisplayInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\auctioneer\Entity\AuctionInterface;
use Drupal\auctioneer\Entity\AuctionTypeInterface;
use Drupal\auctioneer\Entity\BidInterface;
use Drupal\auctioneer\Entity\BidTypeInterface;
use Psr\Log\LoggerInterface;

/**
 * Provides managing service for auction types.
 */
class AuctionTypeManager implements AuctionTypeManagerInterface {

  /**
   * The auction type plugin manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $auctionTypePluginManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The database handler.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The logger factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The current account interface.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $account;

  /**
   * Constructor method.
   *
   * @param \Drupal\Component\Plugin\PluginManagerInterface $auction_type_plugin_manager
   *   The auction type plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Database\Connection $database
   *   The database handler.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Drupal\Core\Session\AccountProxyInterface $account
   *   The current user handler.
   */
  public function __construct(
    PluginManagerInterface $auction_type_plugin_manager,
    EntityTypeManagerInterface $entity_type_manager,
    Connection $database,
    LoggerChannelFactoryInterface $logger_factory,
    AccountProxyInterface $account
  ) {
    $this->auctionTypePluginManager = $auction_type_plugin_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->database = $database;
    $this->loggerFactory = $logger_factory;
    $this->account = $account;
  }

  /**
   * {@inheritdoc}
   */
  public function getAuctionPluginManager() : PluginManagerInterface {
    return $this->auctionTypePluginManager;
  }

  /**
   * {@inheritdoc}
   */
  public function getLogger(string $channel) : LoggerInterface {
    return $this->loggerFactory->get($channel);
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypeManager() : EntityTypeManagerInterface {
    return $this->entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailableAuctionTypeNames($machine_name = FALSE) : array {
    $options = [];
    foreach ($this->auctionTypePluginManager->getDefinitions() as $plugin_definition) {
      $options[$plugin_definition['id']] = $machine_name ? $plugin_definition['id'] : $plugin_definition['label'];
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailableBidTypeNames($machine_name = FALSE) : array {
    $options = [];
    $auction_type_names = $this->getAvailableAuctionTypeNames($machine_name);
    if (!empty($auction_type_names)) {
      $bid_types = $this->entityTypeManager->getStorage('bid_type')->loadMultiple();
      foreach ($bid_types as $bid_type) {
        $group_key = $machine_name ? $bid_type->getPluginName() : $auction_type_names[$bid_type->getPluginName()];
        // Let's group on a common set of items per auction type.
        $options[$group_key][$bid_type->id()] = $machine_name ? $bid_type->id() : $bid_type->label();
      }
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function initializeAuctionTypeBaseConfigurations(string $bid_type_id, AuctionTypeInterface $auction_type, bool $lock_fields = TRUE) : AuctionTypeInterface {
    $entity_type_id = 'auction';
    $bundle = $auction_type->id();
    $bid_type = $this->entityTypeManager->getStorage('bid_type')->load($bid_type_id);
    if ($bid_type instanceof BidTypeInterface) {
      $auction_plugin_id = $bid_type->getPluginName();
      if (!$auction_plugin_id) {
        return $auction_type;
      }
      $plugin_instance = $this->auctionTypePluginManager->createInstance($auction_plugin_id);
      if ($plugin_instance) {
        $fields = $plugin_instance->getAuctionBaseFieldDefinitions();
        foreach ($fields as $field_name => $field_definition) {
          $field_storage = FieldStorageConfig::loadByName($entity_type_id, $field_name);
          if (empty($field_storage)) {
            $field_storage = FieldStorageConfig::create([
              'field_name' => $field_name,
              'entity_type' => $entity_type_id,
              'type' => $field_definition->getType(),
              'cardinality' => $field_definition->getCardinality(),
              'settings' => $field_definition->getSettings(),
              'translatable' => $field_definition->isTranslatable(),
              'locked' => $lock_fields,
            ]);
            $field_storage->save();
          }
          $field_config = FieldConfig::loadByName($entity_type_id, $bundle, $field_name);
          if (empty($field_config)) {
            $field_config = FieldConfig::create([
              'field_storage' => $field_storage,
              'bundle' => $bundle,
              'label' => $field_definition->getLabel(),
              'required' => $field_definition->isRequired(),
              'settings' => $field_definition->getSettings(),
              'translatable' => $field_definition->isTranslatable(),
              'default_value' => $field_definition->getDefaultValueLiteral(),
              'default_value_callback' => $field_definition->getDefaultValueCallback(),
            ]);
            $field_config->save();
          }
          // Show the field on default entity displays, if specified.
          if ($form_display_options = $field_definition->getDisplayOptions('form')) {
            $form_display = $this->getEntityDisplay($entity_type_id, $bundle, 'form');
            $form_display->setComponent($field_name, $form_display_options);
            $form_display->save();
          }
          if ($view_display_options = $field_definition->getDisplayOptions('view')) {
            $view_display = $this->getEntityDisplay($entity_type_id, $bundle, 'view');
            $view_display->setComponent($field_name, $view_display_options);
            $view_display->save();
          }
        }
        $auction_type->setBidTypeName($bid_type_id);
        $auction_type->save();
      }
    }

    return $auction_type;
  }

  /**
   * {@inheritdoc}
   */
  public function initializeBidTypeBaseConfigurations($auction_plugin_id, BidTypeInterface $bid_type, $lock_fields = TRUE) : BidTypeInterface {
    $plugin_instance = $this->auctionTypePluginManager->createInstance($auction_plugin_id);
    if ($plugin_instance) {
      $fields = $plugin_instance->getBidBaseFieldDefinitions();
      $entity_type_id = 'bid';
      $bundle = $bid_type->id();
      foreach ($fields as $field_name => $field_definition) {
        $field_storage = FieldStorageConfig::loadByName($entity_type_id, $field_name);
        if (empty($field_storage)) {
          $field_storage = FieldStorageConfig::create([
            'field_name' => $field_name,
            'entity_type' => $entity_type_id,
            'type' => $field_definition->getType(),
            'cardinality' => $field_definition->getCardinality(),
            'settings' => $field_definition->getSettings(),
            'translatable' => $field_definition->isTranslatable(),
            'locked' => $lock_fields,
          ]);
          $field_storage->save();
        }
        $field_config = FieldConfig::loadByName($entity_type_id, $bundle, $field_name);
        if (empty($field_config)) {
          $field_config = FieldConfig::create([
            'field_storage' => $field_storage,
            'bundle' => $bundle,
            'label' => $field_definition->getLabel(),
            'required' => $field_definition->isRequired(),
            'settings' => $field_definition->getSettings(),
            'translatable' => $field_definition->isTranslatable(),
            'default_value' => $field_definition->getDefaultValueLiteral(),
            'default_value_callback' => $field_definition->getDefaultValueCallback(),
          ]);
          $field_config->save();
        }
        // Show the field on default entity displays, if specified.
        if ($form_display_options = $field_definition->getDisplayOptions('form')) {
          $form_display = $this->getEntityDisplay($entity_type_id, $bundle, 'form');
          $form_display->setComponent($field_name, $form_display_options);
          $form_display->save();
        }
        if ($view_display_options = $field_definition->getDisplayOptions('view')) {
          $view_display = $this->getEntityDisplay($entity_type_id, $bundle, 'view');
          $view_display->setComponent($field_name, $view_display_options);
          $view_display->save();
        }
      }
      $bid_type->setPluginName($auction_plugin_id);
      $bid_type->save();
    }

    return $bid_type;
  }

  /**
   * {@inheritdoc}
   */
  public static function getAllowedAuctionStates() : array {
    $states = [
      'opened' => t('Opened'),
      'paused' => t('Paused'),
      'closed' => t('Closed'),
    ];

    return $states;
  }

  /**
   * {@inheritdoc}
   */
  public function processBidForm(array &$form, FormStateInterface $form_state) {
    /* @var \Drupal\auctioneer\Entity\BidInterface $bid */
    $bid = $form_state->getFormObject()->getEntity();
    // Plugin instance and validation.
    $auction_plugin_id = $bid->getBidType()->getPluginName();
    $plugin_instance = $this->auctionTypePluginManager->createInstance($auction_plugin_id);
    if ($plugin_instance) {
      $plugin_instance->processBidForm($form, $form_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validateBidForm(array &$form, FormStateInterface $form_state) {
    /* @var \Drupal\auctioneer\Entity\BidInterface $bid */
    $bid = $form_state->getFormObject()->getEntity();
    // Plugin instance and validation.
    $auction_plugin_id = $bid->getBidType()->getPluginName();
    $plugin_instance = $this->auctionTypePluginManager->createInstance($auction_plugin_id);
    if ($plugin_instance) {
      $plugin_instance->validateBidForm($form, $form_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitBidForm(array &$form, FormStateInterface $form_state) {
    /* @var \Drupal\auctioneer\Entity\BidInterface $bid */
    $bid = $form_state->getFormObject()->getEntity();
    // Plugin instance and validation.
    $auction_plugin_id = $bid->getBidType()->getPluginName();
    $plugin_instance = $this->auctionTypePluginManager->createInstance($auction_plugin_id);
    if ($plugin_instance) {
      $plugin_instance->submitBidForm($form, $form_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setAuctionHammerBids(array $bids, bool $close_auction = TRUE) {
    $auctions = [];
    // First: Hammer bids. Extract auctions to be closed.
    $uid = $this->account->id();
    $merge = $this->database->merge('auctioneer_auction_hammer');
    foreach ($bids as $bid) {
      $merge->key('bid', $bid->id())->fields([
        'bid' => $bid->id(),
        'uid' => $uid,
        'timestamp' => time(),
      ]);
      /** @var \Drupal\auctioneer\Entity\AuctionInterface $auction */
      $auction = $bid->getAuction();
      $auctions[$auction->id()] = $auction;
    }
    $merge->execute();
    // Second: Close auctions.
    if (!empty($auctions) && $close_auction) {
      foreach ($auctions as $auction) {
        if ($auction->getAuctionState() != 'closed') {
          $auction->set('auction_state', 'closed');
          $auction->save();
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getAuctionHammerBidData(BidInterface $bid) : array {
    $result = $this->database->select('auctioneer_auction_hammer', 'h')
      ->fields('h')
      ->condition('bid', $bid->id())
      ->execute()
      ->fetchAssoc();

    return is_array($result) ? $result : [];
  }

  /**
   * {@inheritdoc}
   */
  public function deleteAuctionHammerBidData(BidInterface $bid) {
    $this->database->delete('auctioneer_auction_hammer')
      ->condition('bid', $bid->id())
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function purgeAuctionHammerBids(AuctionInterface $auction) {
    $hammer = $this->getAuctionHammerBids($auction, FALSE);
    if (!empty($hammer)) {
      $this->database->delete('auctioneer_auction_hammer')
        ->condition('bid', $hammer, 'IN')
        ->execute();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function purgeOrphanBids(array $configuration = []) {
    // Some needed variables to make our query more flexible.
    $limit = $configuration['limit'] ?? 30;
    // Our query itself.
    $query = $this->database->select('bid_field_data', 'bfd');
    $query->fields('bfd', ['id']);
    $query->leftJoin('auction_field_data', 'afd', 'bfd.auction_id = afd.id');
    $query->isNull('afd.id');
    $query->orderBy('bfd.id', 'ASC');
    $query->range(0, $limit);
    $result = $query->execute()->fetchCol();
    // Now, let's delete bids.
    $bid_storage = $this->entityTypeManager->getStorage('bid');
    $orphan_bids = !empty($result) ? $bid_storage->loadMultiple($result) : [];
    if (!empty($orphan_bids)) {
      $bid_storage->delete($orphan_bids);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getAuctionHammerBids(AuctionInterface $auction, bool $load = FALSE) : array {
    // We need to query for bids into the related tables.
    $query = $this->database->select('bid_field_data', 'bfd');
    $query->fields('bfd', ['id']);
    $query->innerJoin('auctioneer_auction_hammer', 'h', 'h.bid = bfd.id');
    $query->condition('bfd.auction_id', $auction->id());
    $result = $query->execute()->fetchCol();

    return $load && !empty($result) ? $this->entityTypeManager->getStorage('bid')->loadMultiple(array_values($result)) : $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getBidTypePluginDefinition(BidTypeInterface $bid_type) : array {
    $plugin_definition = $this->auctionTypePluginManager->getDefinition($bid_type->getPluginName());

    return is_array($plugin_definition) ? $plugin_definition : [];
  }

  /**
   * Get an entity display for the given entity type and bundle.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The bundle.
   * @param string $display_context
   *   The display context ('view' or 'form').
   *
   * @throws \InvalidArgumentException
   *   Thrown when an invalid display context is provided.
   *
   * @return \Drupal\Core\Entity\Display\EntityDisplayInterface
   *   The entity display.
   */
  public function getEntityDisplay($entity_type, $bundle, $display_context) : EntityDisplayInterface {
    if (!in_array($display_context, ['view', 'form'])) {
      throw new \InvalidArgumentException(sprintf('Invalid display_context %s passed. It must be "view" or "form".', $display_context));
    }
    $storage = $this->entityTypeManager->getStorage('entity_' . $display_context . '_display');
    $display = $storage->load($entity_type . '.' . $bundle . '.default');
    if (!$display) {
      $display = $storage->create([
        'targetEntityType' => $entity_type,
        'bundle' => $bundle,
        'mode' => 'default',
        'status' => TRUE,
      ]);
    }

    return $display;
  }

}
