<?php

namespace Drupal\auctioneer\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\auctioneer\AuctionTypeManagerInterface;
use Drupal\auctioneer\Entity\AuctionInterface;
use Drupal\auctioneer\Entity\BidInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the form to mark a bid as winner.
 */
class HammerBidForm extends FormBase {

  /**
   * The auction type manager.
   *
   * @var \Drupal\auctioneer\AuctionTypeManagerInterface
   */
  protected $auctionTypeManager;

  /**
   * Constructs the "HammerBidForm" form.
   *
   * @param \Drupal\auctioneer\AuctionTypeManagerInterface $auction_type_manager
   *   The auction type manager.
   */
  public function __construct(AuctionTypeManagerInterface $auction_type_manager) {
    $this->auctionTypeManager = $auction_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('auctioneer.auction_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    // Workaround for multiple form instances.
    // https://www.drupal.org/project/drupal/issues/2821852#comment-13117654.
    static $count = 0;

    return 'auctioneer_hammer_bid_form_' . $count++;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $build_info = $form_state->getBuildInfo();
    /** @var \Drupal\auctioneer\Entity\BidInterface $bid */
    $bid = auctioneer_get_children($build_info, 'args->0->bid');
    if ($bid instanceof BidInterface) {
      /** @var \Drupal\auctioneer\Entity\AuctionInterface $auction */
      $auction = $bid->getAuction();
      // Setting bid object globally.
      $form_state->set('bid', $bid);
      // Other needed elements.
      $plugin_definition = $this->auctionTypeManager->getBidTypePluginDefinition($bid->getBidType());
      $hammer = $this->auctionTypeManager->getAuctionHammerBidData($bid);
      $form['actions'] = [
        '#type' => 'actions',
      ];
      if (!empty($hammer)) {
        $form['actions']['unhammer'] = [
          '#type' => 'submit',
          '#value' => $this->t('Unset as winner'),
          '#op' => 'unhammer',
        ];
      }
      else {
        $existing_hammer_bids = $auction ? $this->auctionTypeManager->getAuctionHammerBids($auction, FALSE) : [];
        $count_existing_hammers = count($existing_hammer_bids);
        if (!$auction instanceof AuctionInterface) {
          $form['hammer_warning'] = [
            '#type' => 'html_tag',
            '#tag' => 'p',
            '#value' => $this->t('Auction is no longer available.'),
            '#attributes' => [
              'class' => ['hammer-missing-auction'],
            ],
          ];
        }
        elseif ($count_existing_hammers < $plugin_definition['hammer_count']) {
          $form['override'] = [
            '#type' => 'checkbox',
            '#title' => $this->t('Override'),
            '#description' => $this->t('Replace previously existing hammer bids with this one.'),
            '#access' => $count_existing_hammers > 0,
          ];
          $form['actions']['hammer'] = [
            '#type' => 'submit',
            '#value' => $this->t('Set this bid as winner'),
            '#op' => 'hammer',
          ];
        }
        else {
          $form['hammer_warning'] = [
            '#type' => 'html_tag',
            '#tag' => 'p',
            '#value' => $this->t('Auction has reached maximun amount of hammer bids.'),
            '#attributes' => [
              'class' => ['hammer-count-reached'],
            ],
          ];
        }
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);
    $values = $form_state->getValues();
    $bid = $form_state->get('bid');
    $triggering = $form_state->getTriggeringElement();
    if ($bid instanceof BidInterface) {
      $plugin_definition = $this->auctionTypeManager->getBidTypePluginDefinition($bid->getBidType());
      $existing_hammers = $this->auctionTypeManager->getAuctionHammerBids($bid->getAuction(), FALSE);
      if (isset($triggering['#op']) && $triggering['#op'] == 'hammer') {
        if (!isset($values['override']) && (count($existing_hammers) >= $plugin_definition['hammer_count'])) {
          $form_state->setErrorByName('override', $this->t('There is already a winner bid for this auction.'));
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $triggering = $form_state->getTriggeringElement();
    $bid = $form_state->get('bid');
    if ($bid instanceof BidInterface) {
      $plugin_definition = $this->auctionTypeManager->getBidTypePluginDefinition($bid->getBidType());
      $existing_hammers = $this->auctionTypeManager->getAuctionHammerBids($bid->getAuction(), FALSE);
      /** @var \Drupal\auctioneer\Entity\AuctionInterface $auction */
      $auction = $bid->getAuction();
      if (isset($triggering['#op']) && $triggering['#op'] == 'hammer') {
        if (isset($values['override']) || (bool) $values['override']) {
          $this->auctionTypeManager->purgeAuctionHammerBids($auction);
        }
        // Needed to close auction, if hammer count is reached.
        // We add 1, because we're counting current bid.
        $close_auction = (count($existing_hammers) + 1) >= $plugin_definition['hammer_count'];
        $this->auctionTypeManager->setAuctionHammerBids([$bid], $close_auction);
        $this->messenger()->addMessage($this->t('Bid was successfully marked as winner!'));
      }
      else {
        $this->auctionTypeManager->deleteAuctionHammerBidData($bid);
        $this->messenger()->addMessage($this->t('Bid was successfully unmarked as winner!'));
      }
    }
  }

}
